"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class App(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    FLAGS_FIELD_NUMBER: builtins.int
    HASH_FIELD_NUMBER: builtins.int
    HASHCODEDATA_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    flags: builtins.int
    hash: builtins.bytes
    hashCodeData: builtins.bytes
    name: typing.Text
    def __init__(self,
        *,
        flags: builtins.int = ...,
        hash: builtins.bytes = ...,
        hashCodeData: builtins.bytes = ...,
        name: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["flags",b"flags","hash",b"hash","hashCodeData",b"hashCodeData","name",b"name"]) -> None: ...
global___App = App

class AppList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    LIST_FIELD_NUMBER: builtins.int
    @property
    def list(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___App]: ...
    def __init__(self,
        *,
        list: typing.Optional[typing.Iterable[global___App]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["list",b"list"]) -> None: ...
global___AppList = AppList
