"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Parameter(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    ALIAS_FIELD_NUMBER: builtins.int
    LOCAL_FIELD_NUMBER: builtins.int
    name: typing.Text
    alias: typing.Text
    local: builtins.bool
    def __init__(self,
        *,
        name: typing.Text = ...,
        alias: typing.Text = ...,
        local: builtins.bool = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["alias",b"alias","local",b"local","name",b"name"]) -> None: ...
global___Parameter = Parameter

class Request(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    PARAMETERS_FIELD_NUMBER: builtins.int
    REFERENCE_FIELD_NUMBER: builtins.int
    ELF_FIELD_NUMBER: builtins.int
    CLOSE_FIELD_NUMBER: builtins.int
    LARGESTACK_FIELD_NUMBER: builtins.int
    REMOTE_PARAMETERS_FIELD_NUMBER: builtins.int
    id: typing.Text
    parameters: builtins.bytes
    reference: typing.Text
    elf: builtins.bytes
    close: builtins.bool
    largeStack: builtins.bool
    @property
    def remote_parameters(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Parameter]: ...
    def __init__(self,
        *,
        id: typing.Text = ...,
        parameters: builtins.bytes = ...,
        reference: typing.Text = ...,
        elf: builtins.bytes = ...,
        close: builtins.bool = ...,
        largeStack: builtins.bool = ...,
        remote_parameters: typing.Optional[typing.Iterable[global___Parameter]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["close",b"close","elf",b"elf","id",b"id","largeStack",b"largeStack","parameters",b"parameters","reference",b"reference","remote_parameters",b"remote_parameters"]) -> None: ...
global___Request = Request

class Response(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    RESPONSE_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    EXCEPTION_FIELD_NUMBER: builtins.int
    id: typing.Text
    response: builtins.bytes
    message: typing.Text
    exception: typing.Text
    def __init__(self,
        *,
        id: typing.Text = ...,
        response: builtins.bytes = ...,
        message: typing.Text = ...,
        exception: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["exception",b"exception","id",b"id","message",b"message","response",b"response"]) -> None: ...
global___Response = Response
